@make(ais)
@aiInfo(topic=<A static expression cannot have a generic formal type>,
number=[AI-00190/06],
classcode=[BI],
statuscode=[WJ],
mainref=[04.09@\(02)],
date=[89-04-14])
@aiHeading()
@begin(history)
@begin(toplines)
@mainref()
@standard[04.03.02@\(03)]
!class binding interpretation 84-03-13
!status approved by WG9/AJPO 86-07-22
@end(toplines)
!status approved by Director, AJPO 86-07-22
!status approved by WG9/Ada Board 86-07-22
!status approved by Ada Board 86-07-22
!status WG14/ADA Board approved 84-11-27
!status WG14-approved 84-11-27
!status board-approved 84-11-26
!status committee-approved 84-06-28
!status work-item 84-05-24
!status received 84-03-13
@references[!references @CommentNum[83-00286]83-00286, @CommentNum[83-00376]83-00376, @CommentNum[83-00378]83-00378, @CommentNum[83-00379]83-00379, @CommentNum[83-00454]83-00454]
@topic{!topic A static expression cannot have a generic formal type}
@end(history)

@subheading[!summary 84-09-03]

A static expression is not allowed to have a generic formal type (including a
type derived from a generic formal type, directly or indirectly).
(Consequently, if an array's index subtype is a generic formal type,
aggregates for that dimension of the array can have only a single component
association and this component association must have a single choice.)

@subheading[!question 84-09-03]

Consider this example:
@begin(example)
@key[generic]
     @key[type] T @key[is] @key[range] <>;
@key[package] PACK @key[is]
     @key[type] ARR @key[is] @key[array] (T) @key[of] BOOLEAN;
     X : ARR := (1_@|000_@|000 => FALSE,
                 1_@|000_@|001 => TRUE);     -- legal?
@key[end] PACK;
@end(example)
4.3.2(3) requires that when more than one component association is present in
an array aggregate, all the choices must be static.  Consider the aggregate
(1_@|000_@|000 => FALSE, 1_@|000_@|001 => TRUE).  This aggregate is illegal if
1_@|000_@|001 does not belong to the index base type because evaluation of the
literal (i.e., its implicit conversion to the index base type) will raise an
exception [4.9(2)].  If the index base type is a generic formal type, it is
not known at compile time whether 1_@|000_@|001 belongs to the base type or not
(as long as 1_@|000_@|001 does not exceed SYSTEM.MAX_@|INT).  If 1_@|000_@|001 belongs
to the base type for an instantiation, no exception will be raised.  Does the
use of such an aggregate in a generic unit mean the legality of the unit's
instantiation depends on the range of the base type used in the instantiation?

@subheading[!recommendation 84-09-03]

A static expression is not allowed to have a generic formal type (including a
type derived from a generic formal type, directly or indirectly).

@subheading[!discussion 84-09-03]

4.9 specifies the conditions under which an expression can be considered to be
static.  If an expression has a formal generic type, it can only be static
if the formal type is an integer or real type and the primaries in the
expression are either numeric literals, named numbers, function calls as
specified in 4.9(7), or static expressions enclosed in parentheses.  There are
no other possibilities.  In particular, there are no enumeration literals for
a formal discrete type, and since 4.9(11) says a generic formal type is not
static, no constants satisfying 4.9(6) can be declared; moreover, there are no
static attributes for formal generic types, and a formal generic type cannot
be used in a static qualified expression.

In short, an expression such as 1 + 2 or 1.0 + 2.0 can apparently have a
generic formal type and still be considered static according to 4.9.  Can such
expressions be used anywhere static expressions are required?  An examination
of the Standard shows that rules requiring static expressions are given in the
following sections: 3.2.2(1), 3.5.4(3), 3.5.7(3), 3.5.9(3), 3.6.2(2),
3.7.3(3), 4.3.1(2), 4.3.2(3), 4.9(6, 11), 5.4(4), 9.8(1), 13.2(5, 12), and
13.3(4).  These rules generally exclude static expressions that have a generic
formal type because:
@begin(enumerate0)
A  static  expression  is  explicitly  required  to  have   a
universal type: 3.2.2(1), 3.6.2(2), 13.3(3,4);


The type of a static expression is explicitly required to  be
INTEGER: 9.8(1);


An expression that must be  static must have some integer  or
real type that  is determined independent  of the context  of
the expression.  Since universal integer or universal real is
always acceptable in  such contexts,  expressions having  the
form 1 + 2  or 1.0 + 2.0  can never be  considered to have  a
non-universal type  [4.6(15], and  so cannot  have a  generic
formal type: 3.5.4(3), 3.5.7(3), 3.5.9(3), 13.2(5, 12);


Expressions having a generic formal type are excluded because
a generic formal type is not static: 4.9(6, 11);


Generic formal types are explicitly excluded as the type of a
discriminant or as the expression in a case statement, and so
expressions having such types are  not allowed as choices  in
variant records, as discriminants of record aggregates, or as
choices in case statements: 3.7.3(3), 4.3.1(2), 5.4(4).
@end(enumerate0)
The only case remaining, choices in array aggregates [4.3.2(3)], has no
explicit or implicit exclusion regarding generic formal types.  However, since
generic formal types are explicitly excluded by 3.7.3(3) and 5.4(4), and since
these are cases where allowing a generic formal type would mean the legality
of a generic instantiation would depend on the subtype used in its
instantiation, it is reasonable to conclude that the intent was to ensure the
legality of array aggregates would also be determined prior to instantiating a
generic unit.

The required legality check can be ensured in two ways: 1) by not allowing a
choice expression in an array aggregate to have a formal generic type; or 2)
by saying that any expression having a generic formal type is not static.  The
latter rule would allow aggregates such as (1_@|000_@|000..1_@|000_@|001 => TRUE),
since when there is a single choice, it is not required to be static
[4.3.2(3)].  If an array had an index type whose base type was a generic
formal type, the first alternative would forbid any choices for that dimension
except @b[others], since @b[others] is not an expression.  It would even forbid
T'@|FIRST..T'@|LAST, where T is the formal generic type.

Since the intent is to ensure instantiations containing aggregates are legal
if the generic template is legal, and since the second alternative is less
restrictive than the first and accomplishes the intent, the second alternative
is chosen as the recommended interpretation.

@cbon()@begin(appendix)
@subheading[!appendix 89-03-18]
@begin(sectionGroup)
@mybar()

!section 04.09    (02) M. Woodger  89-03-18                          83-01275
@CommentNum[83-01275]!version  1983
!topic Errors in AI-00190/05
@end(sectionGroup)
!reference AI-00190/05

1. In the discussion section of this Commentary, the first paragraph
describes the possibilities for an expression of generic formal
type to be static, and excludes (for example) static qualified
expressions.

The second paragraph says  "An examination of the Standard shows that
rules requiring static expressions are given in the following
sections: ... ". But the list given is not complete. For example,
it omits 4.9(9) which concerns qualified expressions.

This list is in fact only a list of the places that are RELEVANT to
the set of possibilities described in the first paragraph.

To correct this, replace "rules requiring" by
"relevant rules requiring", in the second paragraph.

2. At three places in the first paragraph, "formal generic type"
should be replaced by the standard nomenclature (used everywhere else)
which is "generic formal type".

3. In the fifth paragraph of the discussion, item 3., the expression
[4.6(15]  should read  [4.6(15)].

@end(appendix)@cboff()
